"""
For license, see gpl-3.0.txt
"""
import math
import sys

def chunks(l, n):
    for i in range(0, len(list(l)), n):
        yield l[i:i+n]

def color_to_plane_bits(color, depth):
    """returns the bits for a given pixel in a list, lowest to highest plane"""
    result = [0] * depth
    for bit in range(depth):
        if color & (1 << bit) != 0:
            result[bit] = 1
    return result


def extract_planes(im, depth):
    imdata = im.getdata()
    width, height = im.size

    map_words_per_row = int(width / 16)
    if width % 16 > 0:
        map_words_per_row += 1

    # create the converted planar data
    planes = [[0] * (map_words_per_row * height) for _ in range(depth)]
    for y in range(height):
        x = 0
        while x < width:
            # build a word for each plane
            for i in range(min(16, width - x)):
                # get the palette index for pixel (x + i, y)
                color = imdata[y * width + x + i]  # color index
                planebits = color_to_plane_bits(color, depth)
                # now we need to "or" the bits into the words in their respective planes
                wordidx = (x + i) / 16  # word number in current row
                pos = int(y * map_words_per_row + wordidx)  # list index in the plane
                for planeidx in range(depth):
                    if planebits[planeidx]:
                        planes[planeidx][pos] |= (1 << (15 - (x + i) % 16)) # 1 << ((x + i) % 16)
            x += 16
    return planes, map_words_per_row


def interleave_planes(planes, map_words_per_row):
    """transforms a set of bitplanes into a large array of 16-bit
    word rows. each representing a line of an image
    """
    # 1. for each plane generate a list of lists of <map_words_per_row>
    # values
    chunked = list([list(chunks(plane, map_words_per_row)) for plane in planes])

    # 2. for each plane, add the rows one after another to the result list
    num_rows = len(chunked[0])
    result = []
    for i in range(num_rows):
        for chunk in chunked:
            result.append(chunk[i])
    return result

def write_amiga_image(im, outfile, img_name, use_intuition, interleaved, verbose):
    imdata = im.getdata()
    width, height = im.size
    if width % 16 != 0:
        raise Exception("width needs to be a multiple of 16, is: %d" % width)

    # colors is a list of 3-integer lists ([[r1, g1, b1], ...])
    colors = [i for i in chunks([b for b in im.palette.tobytes()], 3)]
    depth = round(math.log(len(colors), 2))
    if depth == 0:
        raise Exception("images with only 1 color can't be handled")
    # fill the missing colors with black entries
    num_missing_colors = 2 ** depth - len(colors)
    colors += [[0, 0, 0]] * num_missing_colors

    planes, map_words_per_row = extract_planes(im, depth)

    if verbose:
        if interleaved:
            print('using interleaved mode')
        print("image width: %d height: %d depth: %d" % (width, height, depth))
    if use_intuition:
        outfile.write('#include <intuition/intuition.h>\n\n')
    imgdata_varname = '%s_data' % img_name
    outfile.write("/* Ensure that this data is within chip memory or you'll see nothing !!! */\n");
    outfile.write('UWORD __chip %s[] = {\n' % imgdata_varname)
    indent = 4

    if interleaved:
        interleaved_data = interleave_planes(planes, map_words_per_row)
        lines = [(' ' * indent + ','.join(['0x%04x' % v for v in row]))
                 for row in interleaved_data]
        outfile.write(',\n'.join(lines))
    else:
        for i, plane in enumerate(planes):
            if i > 0:
                outfile.write(',\n')
            outfile.write(' ' * indent + '// plane %d\n' % i)
            lines = [(' ' * indent) + ','.join(['0x%04x' % i for i in chunk])
                     for chunk in chunks(plane, map_words_per_row)]
            outfile.write(',\n'.join(lines))
    outfile.write('\n};\n\n')

    planepick = 2 ** depth - 1  # currently we always use all of the planes
    if use_intuition:
        outfile.write('struct Image %s = {\n' % img_name)
        outfile.write(' ' * indent + '0, 0, %d, %d, %d, %s,\n' % (width, height, depth, imgdata_varname));
        outfile.write(' ' * indent + '%d, 0, NULL\n' % planepick)  # PlanePick, PlaneOnOff
        outfile.write('};\n\n')

    outfile.write('UWORD %s_colors[%d] = {\n' % (img_name, len(colors)))
    colors4 = []
    for i, color in enumerate(colors):
        r, g, b = color
        r4 = r >> 4
        g4 = g >> 4
        b4 = b >> 4
        colors4.append('0x%x%x%x' % (r4, g4, b4))
    outfile.write((' ' * indent) + ','.join(colors4) + '\n')
    outfile.write('};\n\n')




"""
fdtool.py - simple FD file processing tool.

This is an alternative to fd2pragma, written in Python.
It was written to understand the FD format and its relationship
with programming languages and to take advantage of the strengths
of a scripting language.
"""
import re
import collections

# the format of a function definition
REGEX = re.compile('^([^()]+)\(([^()]*)\)\(([^()]*)\).*$')


class FDState:
    """represents the parser state"""
    def __init__(self, base, public, offset):
        self.base = base
        self.public = public
        self.offset = offset


def process_command(command, state):
    """process a FD command line"""
    if command.startswith('base'):
        state.base = command.split()[1]
    elif command.startswith('bias'):
        state.offset = int(command.split()[1])
    elif command.startswith('private'):
        state.public = False
    elif command.startswith('public'):
        state.public = True
    elif command.startswith('end'):
        pass
    else:
        print("unsupported command: ", command)


def process_fundef(line, state):
    """process a FD function definition"""
    match = REGEX.match(line)
    if match:
        name, params, regs = match.group(1, 2, 3)
        if state.public:
            if len(params) > 0:
                print("-%d -> %s(%s), [%s]" % (state.offset, name, params, regs))
            else:
                print("-%d -> %s()" % (state.offset, name))
        state.offset += 6


def process(input_file):
    """process the specified input file"""
    state = FDState('', False, 0)
    with open(input_file) as infile:
        for line in infile:
            if line.startswith('*'):
                pass
            elif line.startswith('##'):
                process_command(line[2:], state)
            else:
                process_fundef(line.strip(), state)
    print("Done.")




import os
import jinja2
from datetime import date


C_HEADER_TEMPLATE = """#define VERSION     {{version}}
#define REVISION    {{revision}}
#define DATE        "{{date}}"
#define VERS        "{{app_name}} {{version}}.{{revision}}"
#define VSTRING     "{{app_name}} {{version}}.{{revision}} ({{date}})\\r\\n"
#define VERSTAG     "\\0$VER: {{app_name}} {{version}}.{{revision}} ({{date}})"
"""

ASM_INC_TEMPLATE = """VERSION		EQU	{{version}}
REVISION	EQU	{{revision}}

DATE	MACRO
		dc.b '{{date}}'
		ENDM

VERS	MACRO
		dc.b '{{app_name}} {{version}}.{{revision}}'
		ENDM

VSTRING	MACRO
		dc.b '{{app_name}} {{version}}.{{revision}} ({{date}})',13,10,0
		ENDM

VERSTAG	MACRO
		dc.b 0,'$VER: {{app_name}} {{version}}.{{revision}} ({{date}})',0
		ENDM
"""

def bumprev(version, appname):
    # Step 1. check for <appname>_rev.rev, <appname>_rev.h and <appname>_rev.i
    if not appname.endswith('_rev'):
        appname = appname + '_rev'

    rev_file = "%s.rev" % appname
    h_file = "%s.h" % appname
    i_file = "%s.i" % appname

    if os.path.exists(rev_file):
        with open(rev_file) as infile:
            line = infile.readline()
            revision = int(line.strip())
            revision += 1
    else:
        print("bumprev: creating new file \"%s\"" % rev_file)
        revision = 1

    with open(rev_file, "w") as outfile:
        outfile.write("%d" % revision)

    today = date.today()
    date_str = today.strftime("%d.%m.%Y")
    config = {
        "app_name": appname.replace("_rev", ""),
        "version": version,
        "revision": revision,
        "date": date_str
    }
    ch_templ = jinja2.Template(C_HEADER_TEMPLATE)
    with open(h_file, "w") as outfile:
        outfile.write(ch_templ.render(config))

    ai_templ = jinja2.Template(ASM_INC_TEMPLATE)
    with open(i_file, "w") as outfile:
        outfile.write(ai_templ.render(config))

    print("bumprev: bumped %s to revision %d" % (appname, revision))



import os
import re

from pathlib import Path
from setuptools import setup

NAME = 'amigados-utils'
PACKAGES = ['amigados']
DESCRIPTION = 'amigados-utils is a collection of utilities for Amiga system development'
LICENSE = 'BSD'
URI = 'https://github.com/weiju/amigados-utils'
AUTHOR = 'Wei-ju Wu'
VERSION = '0.1.1'

KEYWORDS = ['amiga', 'system', 'development', 'classic', 'ecs', 'aga']

# See trove classifiers
# https://testpypi.python.org/pypi?%3Aaction=list_classifiers

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Software Development :: Libraries :: Python Modules"
    ]
INSTALL_REQUIRES = ['pillow', 'Jinja2']

PACKAGE_DATA = {
    'amigados': []
}

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

if __name__ == '__main__':
    setup(name=NAME,
          description=DESCRIPTION,
          long_description=long_description,
          long_description_content_type='text/markdown',
          license=LICENSE,
          url=URI,
          version=VERSION,
          author=AUTHOR,
          author_email='weiju.wu@gmail.com',
          maintainer=AUTHOR,
          maintainer_email='weiju.wu@gmail.com',
          keywords=KEYWORDS,
          packages=PACKAGES,
          zip_safe=False,
          classifiers=CLASSIFIERS,
          install_requires=INSTALL_REQUIRES,
          include_package_data=True, package_data=PACKAGE_DATA,
          scripts=['bin/amigados-png2image', 'bin/amigados-fdtool',
                   'bin/amigados-dalf', 'bin/amigados-bumprev',
                   'bin/amigados-dir', 'bin/amigados-copy',
                   'bin/amigados-makedir', 'bin/amigados-createdisk',
                   'bin/amigados-delete'])
